Game = {}

local tracks
local obstacles
local goldOres
local boulders
local boulderWarning
local spawner
local firstRun

local function clamp(min, n, max)
	return n < min and min or n > max and max or n
end

local function rsign()
	return math.random() > 0.5 and 1 or -1
end

local function update(obj, dt)
	obj:update(dt)
end

local function draw(obj)
	obj:draw()
end

local function collidingWithPlayer(obj)
	return Player:colliding(obj)
end

local function notCollidingWithPlayer(obj)
	return not collidingWithPlayer(obj)
end

local function resolvePlayerCollision(obj)
	local dx, dy = collidingWithPlayer(obj)
	if dx and dy then
		Player:resolve(obj, dx, dy)
		return true
	end
	return false
end

local function withinBounds(obj)
	return obj.x + obj.width > 0 and obj.y < Window.height
end

local function playerSpawnPosition()
	return 40, -Player.height
end

local function trackSittingPosition(track, sittingObj)
	return math.random(track.x, track.x + track.width - sittingObj.width), track.y - sittingObj.height - 3
end

local function drawRect(rect)
	love.graphics.rectangle('line', rect.x, rect.y, rect.width, rect.height)
end

local function boulderTime()
	return math.random() * 2 + 3
end


function Game:spawnTrack(y)
	assert(y, "must be given a y value for track spawn")
	local halfHeight = Window.height
	local track = new(Track, Window.width, y)
	table.insert(tracks, track)
	table.sort(tracks, function(a,b) return a.y < b.y end)

	for i=1, math.random(0, 3) do
		self:spawnObstacle(track)
	end
	if math.random() < 1/2 then
		self:spawnGold(track)
	end

	return track
end

function Game:spawnObstacle(track)
	local obstacle = new(Obstacle, trackSittingPosition(track, Obstacle))
	obstacle.y = obstacle.y + 3
	table.insert(obstacles, obstacle)
	return obstacle
end

function Game:spawnGold(track)
	local gold = new(Gold, trackSittingPosition(track, Gold))
	gold.y = gold.y - 2
	table.insert(goldOres, gold)
	return gold
end

function Game:spawnBoulder(deviation)
	local x = (Player.x + Player.width / 2) + math.random(deviation) * rsign()
	local boulder = new(Boulder, clamp(0, x, Window.width), -Boulder.radius)

	spawner:add(0.5, function()
		table.insert(boulders, boulder)
	end)

	return boulder
end


function Game:init()
	spawner = Timer()
end

function Game:enter()
	tracks = {}
	goldOres = {}
	obstacles = {}
	boulders = {}

	spawner:clear()

	spawner:add(0, function(spawnTrack)
		local track = self:spawnTrack(math.random(Window.height / 2, Window.height - 32))
		local time = track.width / (16 * 9) - math.random() / 2
		spawner:add(time, spawnTrack)
	end)

	spawner:add(boulderTime(), function(func)
		if Player.alive then
			if math.random() > 1/5 then
				boulderWarning = self:spawnBoulder(30)
				spawner:add(1, function() boulderWarning = nil end)
			else
				for i=1, 3 do
					spawner:add(i * 0.5, function()
						boulderWarning = self:spawnBoulder(60)
					end)
				end
				spawner:add(2.5, function() boulderWarning = nil end)
			end
		end
		spawner:add(boulderTime(), func)
	end)
 
	Player:init()
end

function Game:keypressed(k)
	if k == 'return' or k == ' ' and not Player.alive then
		Player:spawn(playerSpawnPosition())
	end
end

function Game:update(dt)
	-- update player input
	Input:update(dt, Player)

	-- move the player
	Player:update(dt)

	-- remove anything that isn't within the bounds of the window
	sweep(tracks, withinBounds)
	sweep(goldOres, withinBounds)
	sweep(obstacles, withinBounds)
	sweep(boulders, withinBounds)

	-- move the game stuff
	map(goldOres, update, dt)
	map(obstacles, update, dt)
	map(tracks, update, dt)
	map(boulders, update, dt)

	-- resolve collisions with the player
	map(tracks, resolvePlayerCollision)
	map(boulders, resolvePlayerCollision)
	map(obstacles, resolvePlayerCollision)

	-- count gold ore scores with the player, remove colliding ores (keep noncolliding ores)
	sweep(goldOres, inverse(resolvePlayerCollision))

	-- kill the player if he fell
	if Player.y > Window.height and Player:kill() then
		Window:shake(1)
	end

	-- spawn stuff
	spawner:update(dt)
end

function Game:draw()
	Window:draw(function()
		Scenery:drawBackground()

		map(obstacles, draw)
		map(tracks, draw)
		map(goldOres, draw)
		map(boulders, draw)
		Player:draw()

		Scenery:drawForeground()

		love.graphics.setFont(Fonts.numbers)
		love.graphics.printf(Player.score, 0, 5, Window.width, 'center')

		if boulderWarning and GAME_TIME * 5 % 1 > 0.5 then
			love.graphics.draw(Images.arrow, boulderWarning.x, 20)
		end

		local text = string.format("HIGH SCORE: %d\nDEATHS: %d\n\nARROW KEYS TO MOVE\nPRESS ENTER TO BEGIN",
			Player.highScore, Player.deaths)
		if not Player.alive then
			love.graphics.setFont(Fonts.main)
			love.graphics.setColor(0, 0, 0, 120)
			love.graphics.printf(text, 0, Window.height - 50, Window.width, 'center')
			love.graphics.setColor(255, 255, 255)
			love.graphics.printf(text, 0, Window.height - 50 - 1, Window.width, 'center')
		end
	end)
end
