Player = {
	width = 10;
	height = 20;
	angle = 0;
	jumpTime = 0;
	alive = false;
	invincibility = 0;
	score = 0;
	highScore = 0;
	deaths = 0;
}

local grid, animation


local function interpolate(value, target, delta)
	return value - (value - target) * math.min(delta, 1)
end

local function loadNumberFromFile(filepath, default)
	if not love.filesystem.exists(filepath) then
		love.filesystem.write(filepath, tostring(default))
	end
	local content = love.filesystem.read(filepath)
	local num = tonumber(content)
	if not num then
		love.filesystem.write(filepath, tostring(default))
	else
		return num
	end
	return default
end


function Player:load()
	grid = anim8.newGrid(16, 32, Images.player:getDimensions())
	animation = anim8.newAnimation(grid('1-4', 1), 0.05)
	Sounds.trackroll:setLooping(true)
	Sounds.trackroll:setVolume(0.2)

	self.highScore = loadNumberFromFile("highscore", 0)
	self.deaths = loadNumberFromFile("deaths", 0)
end

function Player:init()
	Physics.implement(self)
	self.standingOn = {}
	self.y = Window.height
end

function Player:spawn(x, y)
	if not self.alive then
		self.x, self.y = x, y
		self.alive = true
		self.jumpTime = 0
		self.score = 0
		self.yvel = 0
		self.invincibility = 1
		return true
	end
	return false
end

function Player:kill()
	if self.alive then
		self.alive = false
		self.deaths = self.deaths + 1
		Resources:playSound(Sounds.death)
		love.filesystem.write("highscore", tostring(self.highScore))
		love.filesystem.write("deaths", tostring(self.deaths))
		return true
	end
	return false
end

function Player:update(dt)
	Physics.update(self, dt)
	if self.alive then
		self.jumpTime = self.jumpTime - dt
		self.invincibility = self.invincibility - dt

		local newSpeed = 0
		if Input.movingLeft then
			newSpeed = newSpeed - MOVE_SPEED
		elseif Input.movingRight then
			newSpeed = newSpeed + MOVE_SPEED
		end
		self.xvel = interpolate(self.xvel, newSpeed, dt * 5)

		if Input.jumping then
			self:jump()
		else
			self:stopJump()
		end

		sweep(self.standingOn, function(obj)
			return Physics.colliding(self, obj)
		end)

		if self.x < 0 then
			self.x = 0
		elseif self.x + self.width > Window.width then
			self.x = Window.width - self.width
		end

		self.angle = self.xvel / 400
	else
		self.angle = self.angle - math.pi * 2 * dt
	end

	if self:onGround() then
		if not Sounds.trackroll:isPlaying() then
			Sounds.trackroll:play()
		end
	else
		Sounds.trackroll:stop()
	end

	animation:update(dt)
end

function Player:jump()
	if self.jumpTime <= 0 and self:onGround() then
		self.jumpTime = JUMP_TIME
	end
	if self.jumpTime > 0 then
		self.yvel = -JUMP_SPEED
		self.y = self.y - 1
		if self:onGround() then
			Resources:playSound(Sounds.jump)
		end
	end
end

function Player:stopJump()
	self.jumpTime = 0
end

function Player:onGround()
	return self.alive and #self.standingOn > 0
end

function Player:incScore(n)
	if self.alive then
		self.score = self.score + n
		if self.score > self.highScore then
			self.highScore = self.score
		end
	end
end

function Player:colliding(other)
	if self.alive then
		return Physics.colliding(self, other)
	end
end

function Player:resolve(other, dx, dy)
	if other.__parent == Gold then
		self:incScore(1)
		Resources:playSound(Sounds.coin)
	elseif other.__parent == Track then
		local yvel = self.yvel
		if not Input.droppingDown and Physics.resolvePassThrough(self, other, dx, dy) then
			if not find(self.standingOn, searchFor(other)) then
				table.insert(self.standingOn, other)
				Window:shake(math.min(yvel / 500, 1))
				Sounds.trackhit:setVolume(math.min(yvel / 300, 1))
				Resources:playSound(Sounds.trackhit)
			end
		end
	elseif other.__parent == Obstacle then
		-- make the object "bounce"
	elseif other.__parent == Boulder then
		if self.invincibility < 0 then
			self.yvel = -200
			self:kill()
			Window:shake(1)
		end
	end
end

function Player:draw()
	local dy = 1 + (self:onGround() and GAME_TIME * 15 % 1 or 0)
	local img = Images.player

	if self.invincibility < 0 or GAME_TIME * 20 % 1 > 0.5 then
		animation:draw(img,
			self.x + self.width/2, self.y + self.height / 2 - dy,
			self.angle, 1, 1, 8, 16)
	end
end
